/* file_wrappers.c
 *
 * $Id: file_wrappers.c,v 1.3 2005/10/21 11:41:44 richardv Exp $
 *
 * Wiretap Library
 * Copyright (c) 1998 by Gilbert Ramirez <gram@alumni.rice.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* size of the buffer. must be a power of 2.
 *
 * XXX: what's an appropriate setting for this?
 */
#define BUFFER_SIZE 8192 

/* #define DEBUG_BUF */

/*
 * OK, now this is tricky.
 *
 * At least on FreeBSD 3.2, "/usr/include/zlib.h" includes
 * "/usr/include/zconf.h", which, if HAVE_UNISTD_H is defined,
 * #defines "z_off_t" to be "off_t", and if HAVE_UNISTD_H is
 * not defines, #defines "z_off_t" to be "long" if it's not
 * already #defined.
 *
 * In 4.4-Lite-derived systems such as FreeBSD, "off_t" is
 * "long long int", not "long int", so the definition of "z_off_t" -
 * and therefore the types of the arguments to routines such as
 * "gzseek()", as declared, with prototypes, in "zlib.h" - depends
 * on whether HAVE_UNISTD_H is defined prior to including "zlib.h"!
 *
 * It's not defined in the FreeBSD 3.2 "zlib", so if we include "zlib.h"
 * after defining HAVE_UNISTD_H, we get a misdeclaration of "gzseek()",
 * and, if we're building with "zlib" support, anything that seeks
 * on a file may not work.
 *
 * Other BSDs may have the same problem, if they haven't done something
 * such as defining HAVE_UNISTD_H in "zconf.h".
 *
 * "config.h" defines HAVE_UNISTD_H, on all systems that have it, and all
 * 4.4-Lite-derived BSDs have it.  Therefore, given that "zlib.h" is included
 * by "file_wrappers.h", that means that unless we include "zlib.h" before
 * we include "config.h", we get a misdeclaration of "gzseek()".
 *
 * Unfortunately, it's "config.h" that tells us whether we have "zlib"
 * in the first place, so we don't know whether to include "zlib.h"
 * until we include "config.h"....
 *
 * A similar problem appears to occur with "gztell()", at least on
 * NetBSD.
 *
 * To add further complication, on recent versions, at least, of OpenBSD,
 * the Makefile for zlib defines HAVE_UNISTD_H.
 *
 * So what we do is, on all OSes other than OpenBSD, *undefine* HAVE_UNISTD_H
 * before including "wtap.h" (we need "wtap.h" to get the WTAP_ERR_ZLIB
 * values, and it also includes "zlib.h" if HAVE_ZLIB" is defined), and,
 * if we have zlib, make "raw_file_seek()" and "raw_file_tell()" subroutines, so
 * that the only calls to "gzseek()" and "gztell()" are in this file, which,
 * by dint of the hackery described above, manages to correctly declare
 * "gzseek()" and "gztell()".
 *
 * On OpenBSD, we forcibly *define* HAVE_UNISTD_H if it's not defined.
 *
 * Hopefully, the BSDs will, over time, remove the test for HAVE_UNISTD_H
 * from "zconf.h", so that "gzseek()" and "gztell()" will be declared
 * with the correct signature regardless of whether HAVE_UNISTD_H is
 * defined, so that if they change the signature we don't have to worry
 * about making sure it's defined or not defined.
 *
 * DO NOT, UNDER ANY CIRCUMSTANCES, REMOVE THE FOLLOWING LINES, OR MOVE
 * THEM AFTER THE INCLUDE OF "wtap.h"!  Doing so will cause any program
 * using Wiretap to read capture files to fail miserably on a FreeBSD
 * 3.2 or 3.3 system - and possibly some other BSD systems - if zlib is
 * installed.  If you *must* include <unistd.h> here, do so *before*
 * including "wtap.h", and before undefining HAVE_UNISTD_H.  If you
 * *must* have HAVE_UNISTD_H defined before including "wtap.h", put
 * "raw_file_error()" into a file by itself, which can cheerfully include
 * "wtap.h" and get "gzseek()" misdeclared, and include just "zlib.h"
 * in this file - *after* undefining HAVE_UNISTD_H.
 */
#ifdef __OpenBSD__
#ifndef HAVE_UNISTD_H
#define HAVE_UNISTD_H
#endif /* HAVE_UNISTD_H */
#else /* __OpenBSD__ */
#undef HAVE_UNISTD_H
#endif /* __OpenBSD__ */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "wtap-int.h"
#include "file_wrappers.h"


/*******************************************************************************
 *
 * We start with a collection of routines named raw_foo(), which are
 * essentially wrappers around some basic stdio or zlib functions.
 *
 * Most of them are just macros, but a few of them have a bit more intelligence
 * built in.
 *
 ******************************************************************************/

#ifdef HAVE_LIBZ

#define RAW_FILE void

#define raw_file_open gzopen
#define raw_filed_open gzdopen
#define raw_file_read(buf, count, file) gzread((file),(buf),(count))
static size_t raw_file_write(void *buf, size_t count, RAW_FILE *file);
#define raw_file_close gzclose
#define raw_file_getc gzgetc
#define raw_file_gets(buf, len, file) gzgets((file), (buf), (len))
static int raw_file_error(RAW_FILE *fh);
#define raw_file_eof gzeof
#define raw_file_clearerr gzclearerr
static int raw_file_seek(gzFile stream, off_t offset, int whence, int *err);
#define raw_file_tell(stream) ((off_t)(gztell(stream)))

#else /* No zLib */

#define RAW_FILE FILE

#define raw_file_open fopen
#define raw_filed_open fdopen
#define raw_file_read(buf, count, file) fread((buf),1,(count),(file))
#define raw_file_write(buf, count, file) fwrite((buf),1,(count),(file))
#define raw_file_close fclose
#define raw_file_tell ftell
#define raw_file_getc fgetc
#define raw_file_gets fgets
#define raw_file_eof feof
#define raw_file_clearerr clearerr
static int raw_file_seek(FILE *stream, off_t offset, int whence, int *err);
#ifdef HAVE_FSEEKO
#define raw_file_tell(stream) ftello(stream)
#else
#define raw_file_tell(stream) ((off_t)ftell(stream))
#endif
#endif /* HAVE_LIBZ */

#ifdef HAVE_LIBZ
static int raw_file_seek(gzFile stream, off_t offset, int whence, int *err)
{
	z_off_t ret;

	ret = gzseek(stream, (z_off_t)offset, whence);
	if (ret == -1) {
		/*
		 * XXX - "gzseek()", as of zlib 1.1.4, doesn't set
		 * "z_err" for the stream, so "gzerror()" could return
		 * a bogus Z_OK.
		 *
		 * As this call failed, we know "gzerror()" shouldn't
		 * return Z_OK; if it does, we assume that "errno" is
		 * the real error.
		 */
		*err = raw_file_error(stream);
		if (*err == 0)
			*err = errno;
                return -1;
	}
	return 0;
}

static size_t raw_file_write(void *buf, size_t count, RAW_FILE *file)
{
    /* 
     * gzwrite sometimes returns a negative value on error, it appears, despite
     * what zlib.h might say
     */
    int r = gzwrite(file,buf,count);
    if(r < 0)
        return 0;
    else
        return (size_t)r;
}

#else
static int raw_file_seek(FILE *stream, off_t offset, int whence, int *err)
{
	int ret;

#ifdef HAVE_FSEEKO
	ret = fseeko(stream, offset, whence);
#else
	ret = fseek(stream, (long)offset, whence);
#endif
	if (ret == -1)
		*err = raw_file_error(stream);
	return ret;
}
#endif /* HAVE_LIBZ */

/*
 * Routine to return a Wiretap error code (0 for no error, an errno
 * for a file error, or a WTAP_ERR_ code for other errors) for an
 * I/O stream.
 */
#ifdef HAVE_LIBZ
static int
raw_file_error(gzFile fh)
{
	int errnum;

	gzerror(fh, &errnum);
	switch (errnum) {

	case Z_OK:		/* no error */
		return 0;

	case Z_STREAM_END:	/* EOF - not an error */
		return 0;

	case Z_ERRNO:		/* file I/O error */
		return errno;

	default:
		return WTAP_ERR_ZLIB + errnum;
	}
}
#else /* HAVE_LIBZ */
static int
raw_file_error(FILE *fh)
{
	if (ferror(fh))
		return errno;
	else
		return 0;
}
#endif /* HAVE_LIBZ */



/*******************************************************************************
 *
 * Now we have the implementation of the buffering layer.
 *
 ******************************************************************************/

#if (((BUFFER_SIZE) & ((BUFFER_SIZE)-1))!=0)
#error BUFFER_SIZE must be a power of two
#endif

#ifdef EINVAL
# define MAYBE_SET_EINVAL() errno=EINVAL
#else
# define MAYBE_SET_EINVAL() /* nothing */
#endif

#define idx_clip(idx) ((idx) & ((BUFFER_SIZE)-1))
#define start_ptr(buf) (idx_clip((buf)->start))
#define end_ptr(buf) (idx_clip((buf)->end))
#define ptr_ptr(buf) (idx_clip((buf)->ptr))
#define buf_avail(buf) (((buf)->end)-((buf)->ptr))

/*
 * This is the structure that holds all the information we need to perform
 * buffering on a stream.
 *
 * The buffer itself is a circular buffer - that is to say, its contents simply
 * wrap around, rather than being shuffled around in memory. The indices into
 * the buffer - start, end and ptr - are allowed to increment without limit,
 * and should hence be used modulo BUFFER_SIZE (start_ptr(), end_ptr() and
 * ptr_ptr() are defined to do this for you).
 *
 * We maintain the following invariant for the indices:
 *    start <= ptr <= end
 *
 * It's assumed that the underlying file pointer is always at 'end' - ie, the
 * next byte to be read from the file should go into buf[end % BUFFER_SIZE].
 * Whilst this occasionally leads to the buffer being flushed unnecessarily,
 * the complexity of keeping a separate pointer almost certainly outweighs
 * this, particularly since the purpose of this implementation is not primarily
 * to perform buffering of disk access (that's stdio's job), but to permit
 * seeking in pipes.
 *
 * (Note the index invariant doesn't strictly hold true at the wrap-point of
 * size_t; hence it might be more accurately expressed as
 *    1 ptr - start >= 0
 *    2 end - ptr >= 0)
 */
struct buffered_file {
    RAW_FILE *fh;
 
    size_t start;          /* the index of the first valid byte in the buffer */
    size_t end;            /* the index of the first invalid byte after the
                            * valid data */
    size_t ptr;            /* our own file position indicator: the index of the
                            * next byte in the buffer to be read */

    /* what raw_file_tell() would return now */
    off_t file_offset;
    char buf[BUFFER_SIZE]; /* a circular buffer */
};

#ifdef DEBUG_BUF
static void dump_buf_info(buffered_file *stream)
{
    fprintf(stderr,"    Buf %p: (start, ptr, end): (%#x, %#x, %#x)\n"
                   "                    buffered: %u bytes; available: %u bytes; file_offset: %li\n",
            stream,
            stream->start,
            stream->ptr,
            stream->end,
            stream->end-stream->start,
            buf_avail(stream),
            stream->file_offset);
    g_assert((long)(stream->ptr-stream->start)>=0);
    g_assert((long)(stream->end-stream->ptr)>=0);
}
#else
#endif

static void flush_buf(buffered_file *stream) {
    stream->start = 0;
    stream->end = 0;
    stream->ptr = 0;
}
    

static buffered_file *init_buffered_file(RAW_FILE *fh, const char *mode){

    buffered_file *result;
    off_t i;

    result = malloc(sizeof(buffered_file));
    if(result == NULL)
        return NULL;
    result->fh = fh;
    result->start = 0;
    result->end = 0;
    result->ptr = 0;

    if(strchr(mode,'a')) {
        /* we opened in append mode, sort out the file_offset */
        i = raw_file_tell(fh);
        if(i < 0) {
#ifdef DEBUG_BUF
            fprintf(stderr,"file_open: unable to ftell: %s\n",strerror(errno));
            dump_buf_info(result);
#endif
            free(result);
            raw_file_close(fh);
            return NULL;
        }
        result->file_offset = i;
    } else {
        result->file_offset = 0;
    }
        
#ifdef DEBUG_BUF
    fprintf(stderr,"file_open: success: fh=%p\n",fh);
    dump_buf_info(result);
#endif
    return result;
}

/* ensures that 'len' bytes are available for reading from the buffer. Returns
 * -1 on error, 0 on success
 *
 * we reqiure len <= BUFFER_SIZE
 *
 * if 'seek' is non-zero, and we hit eof, the file is extended and the
 * buffer is filled with zeros.
 * otherwise, we return -1 at eof.
 */
static int ensure_available(buffered_file *stream, size_t len, gboolean seek)
{
    size_t avail = buf_avail(stream);
    size_t delta;
    

#ifdef DEBUG_BUF
    fprintf(stderr,"ensure_available: requires %u bytes\n",len);
    dump_buf_info(stream);
#endif

    g_assert(len <= BUFFER_SIZE);

    if(len <= avail)
        return 0;

    delta = len - avail;
    while(delta > 0) {
        size_t offs = end_ptr(stream);
        size_t nbytes;
        size_t res;
        
        if(offs + delta > BUFFER_SIZE) {
            nbytes = BUFFER_SIZE-offs;
#ifdef DEBUG_BUF
            fprintf(stderr, "ensure_available: on the cusp, trimming initial read to %u bytes\n",
                    nbytes);
#endif
        } else {
            nbytes = delta;
        }

        res = raw_file_read(stream->buf + offs, nbytes, stream -> fh);
        stream -> file_offset += res;
        stream -> end += res;
        delta -= res;
        nbytes -= res;

        /* we may well have blatted over the start of the buffer, so check for
         * that */
        if(stream -> end - stream -> start > BUFFER_SIZE )
            stream -> start = stream -> end - BUFFER_SIZE;

        if(nbytes > 0) {
            if(raw_file_error(stream->fh) || !seek) {
#ifdef DEBUG_BUF
                if(raw_file_error(stream->fh))
                    fprintf(stderr,"ensure_available: unable to read: %s (%d)\n",strerror(errno),errno);
                else
                    fprintf(stderr,"ensure_available: eof\n");

                dump_buf_info(stream);
#endif
                return -1;
            }
            offs += res;
            memset(stream->buf + offs, 0, nbytes);
            stream -> end += res;
            
            res = raw_file_write(stream->buf + offs, nbytes, stream -> fh);
            stream -> file_offset += res;

            if(nbytes > res) {
#ifdef DEBUG_BUF
                fprintf(stderr,"ensure_available: failed to extend file by %u bytes (only wrote %u bytes)\n",
                        nbytes, res);
                dump_buf_info(stream);
#endif
                return -1;
            }
            delta -= res;
#ifdef DEBUG_BUF
            fprintf(stderr,"ensure_available: extended file by %u bytes\n",
                    res);
#endif
                
        }
    }

    avail = buf_avail(stream);
    g_assert(avail >= len);
    
#ifdef DEBUG_BUF
    fprintf(stderr,"ensure_available: request satisfied\n");
    dump_buf_info(stream);
#endif

    return 0;
}


buffered_file *file_open(const char *path, const char *mode)
{
    FILE *fh;

    if((fh = raw_file_open(path,mode)) == NULL)
        return NULL;

    return init_buffered_file(fh,mode);
}


buffered_file *filed_open(int fd, const char *mode)
{
    FILE *fh;

    if((fh = raw_filed_open(fd,mode)) == NULL)
        return NULL;

    return init_buffered_file(fh,mode);
}


int file_seek(buffered_file *stream, off_t offset, int whence, int *err)
{
    off_t abs_pos;
    int avail, ravail;

    /* These casts to int are ok because the values are known
     * to be at most 2*BUFFER_SIZE */
    avail = (int)buf_avail(stream);
    ravail = -(int)(stream->ptr - stream->start);

    if(stream==NULL) {
        MAYBE_SET_EINVAL();
        return -1;
    }
            
    /* start by figuring out where in the file we want to get to,
       and its relative offset to our current pointer */
    switch(whence) {
        case SEEK_SET:
            abs_pos = offset;
            offset -= stream->file_offset;
            offset += avail;
            break;
        case SEEK_CUR:
            abs_pos = offset - avail;
            abs_pos += stream -> file_offset;
            break;
        case SEEK_END:
            /* we don't know where the end of the file is, so this is hard to
             * implement and isn't actually used. */
        default:
            g_assert_not_reached();
            abs_pos = 0;
            break;
    }

#ifdef DEBUG_BUF
    fprintf(stderr,"file_seek: seeking to position %li; offset from ptr %li\n",abs_pos,offset);
    dump_buf_info(stream);
#endif

    /* if we need to go backwards, we just seek to the relevant place and flush
     * the buffer. */
    if(ravail > offset) {
#ifdef DEBUG_BUF
        fprintf(stderr,"file_seek: reverse seek out of range, doing raw seek\n");
#endif

        flush_buf(stream);
        offset -= avail; /* make the offset relative to the file ptr, not the
                          * public ptr, now that we've flushed the data between them */

        if(raw_file_seek(stream->fh,offset,SEEK_CUR,err) < 0) {
#ifdef DEBUG_BUF
            fprintf(stderr,"file_seek: raw seek failed: %d,%s(%d)\n",err,strerror(errno),errno);
#endif
            return -1;
        }
        stream->file_offset=abs_pos;
#ifdef DEBUG_BUF
        fprintf(stderr,"file_seek: success\n");
        dump_buf_info(stream);
#endif
        return 0;
    }
    

    /* if it's miles off the end of the buffer, we read, and discard,
     * bucketfuls of data. */
    if(avail + BUFFER_SIZE <= offset) {
        long nbytes;
        
#ifdef DEBUG_BUF
        fprintf(stderr,"file_seek: long forward seek: doing data discard\n");
#endif
        flush_buf(stream);
        offset -= avail; /* make the offset relative to the file ptr, not the
                          * public ptr, now that we've flushed the data between them */

        /* we attempt to keep the buffer as full as possible, so actually seek
         * to BUFFER_SIZE bytes before where we want
         * to be and then fall out to the normal path. */
        nbytes = offset-BUFFER_SIZE;
        
        while(nbytes > 0) {
            size_t nb = nbytes < BUFFER_SIZE ? nbytes : BUFFER_SIZE;
            if( ensure_available(stream,nb,1) != 0 ) {
#ifdef DEBUG_BUF
                fprintf(stderr,"file_seek: ensure_available failed\n");
#endif
                return -1;
            }
            flush_buf(stream);
            nbytes -= nb;
        }

        offset = BUFFER_SIZE;
    }


    if( offset > 0 ) {
        if( ensure_available(stream,offset,1) != 0 ) {
#ifdef DEBUG_BUF
            fprintf(stderr,"file_seek: ensure_available failed\n");
#endif
            return -1;
        }
    }


    /* we now know that we're seeking within the buffer, so we can just update
     * offset */
#ifdef DEBUG_BUF
    if(offset < 0)
        g_assert(stream->ptr >= -offset);
#endif
    stream->ptr += offset;

#ifdef DEBUG_BUF
    fprintf(stderr,"file_seek: success\n");
    dump_buf_info(stream);
#endif

    return 0;

}

size_t file_read(void *ptr, size_t size, size_t nmemb, buffered_file *stream)
{
    size_t delta = size*nmemb;
    
    if(stream==NULL) {
        MAYBE_SET_EINVAL();
        return 0;
    }

    while(delta > 0) {
        size_t nbytes = delta;
        size_t avail;
        
        if(nbytes > BUFFER_SIZE) {
            nbytes = BUFFER_SIZE;
#ifdef DEBUG_BUF
            fprintf(stderr, "file_read: huge read, trimming initial buffer fill to %u bytes\n",
                    nbytes);
#endif
        }

        ensure_available(stream,nbytes,0);

        avail = buf_avail(stream);
        if(avail < nbytes) {
            /* ensure_available failed. We should still return all we can, but
             * not loop hereafter */
            delta = nbytes = avail;
            nmemb = avail/size;
        }

        while(nbytes > 0) {
            size_t offs = ptr_ptr(stream);
            size_t len;
            
            if(offs + nbytes > BUFFER_SIZE) {
                len = BUFFER_SIZE-offs;
#ifdef DEBUG_BUF
                fprintf(stderr, "read: on the cusp, trimming initial read to %u bytes\n",
                        len);
#endif
            } else {
                len = nbytes;
            }

            memcpy(ptr, stream -> buf + offs, len);
            ptr = ((char *)ptr)+len;
            nbytes -= len;
            delta -= len;
            stream -> ptr += len;
        }
    }

    return nmemb;
}


#if 0
/* this isn't used, so hasn't been tested. It's believed correct, however. */
size_t file_write(void *ptr, size_t size, size_t nmemb, buffered_file *stream)
{
    size_t nbytes;
    int avail;
    
    if(stream==NULL) {
        MAYBE_SET_EINVAL();
        return 0;
    }


    if(nmemb == 0 || size == 0) {
        return nmemb;
    }

    /* if our ptr isn't at the end of the buffer, trim the end of the
     * buffer */
    avail = (int)buf_avail(stream);
    if(avail > 0) {
        int err;
        
#ifdef DEBUG_BUF
        fprintf(stderr,"file_write: trimming end of buffer\n");
        dump_buf_info(stream);
#endif
        if(raw_file_seek(stream->fh,-avail,SEEK_CUR,&err)!= 0) {
#ifdef DEBUG_BUF
            fprintf(stderr,"file_write: raw seek failed: err %i\n",err);
#endif
            return 0;
        }
        stream->end = stream->ptr;
    }

    /* first do the write */
    nbytes = raw_file_write(ptr,size*nmemb,stream->fh);

    /* then update the buffer */
    if(nbytes > BUFFER_SIZE) {
        /* only copy the end of the data if we did a huge write */
        ptr = ((char *)ptr)+BUFFER_SIZE-nbytes;
        nbytes = BUFFER_SIZE;
    }
        
    while(nbytes > 0) {
        size_t offs = ptr_ptr(stream);
        size_t len;
            
        if(offs + nbytes > BUFFER_SIZE) {
            len = BUFFER_SIZE-offs;
#ifdef DEBUG_BUF
            fprintf(stderr, "write: on the cusp, trimming initial read to %u bytes\n",
                    len);
#endif
        } else {
            len = nbytes;
        }

        memcpy(stream -> buf + offs, ptr, len);
        ptr = ((char *)ptr)+len;
        nbytes -= len;
        stream -> ptr += len;
    }

    return nbytes/size;
}
#endif

int file_close(buffered_file *stream)
{
    int r;
#ifdef DEBUG_BUF
    fprintf(stderr,"file_close: fh=%p\n",stream->fh);
    dump_buf_info(stream);
#endif
    r= raw_file_close(stream->fh);
    stream->fh=NULL;
    free(stream);
    return r;
}

off_t file_tell(buffered_file *stream)
{
    if(stream==NULL) {
        MAYBE_SET_EINVAL();
        return 0;
    }
    return stream -> file_offset;
}

int file_getc(buffered_file *stream)
{
    int r;
    
    if(stream==NULL) {
        MAYBE_SET_EINVAL();
        return 0;
    }

    if(ensure_available(stream,1,0) == -1)
        return EOF;

    r = (int)stream->buf[ptr_ptr(stream)];
    stream->ptr++;
    return r;
}

char *file_gets(char *s, int size, buffered_file *stream)
{
    char *ptr = s;

    while(size > 1) {
        size_t nbytes = (size > BUFFER_SIZE) ? BUFFER_SIZE : size-1;
        size_t idx;
        
        if(ensure_available(stream,nbytes,0) != 0) {
            if(raw_file_error(stream->fh))
                return NULL;
            /* must be eof. clip the read to the amount of data available,
               and don't loop hereafter. */
            nbytes = buf_avail(stream);
            size = nbytes + 1;
        }

        idx = ptr_ptr(stream);
        size -= nbytes;
        while(nbytes > 0) {
            *ptr = stream->buf[idx];
            idx = idx_clip(idx+1);
            stream->ptr++;
            if(*ptr++ == '\n') {
                size = 0;
                break;
            }
            nbytes--;
        }
    }

    if(ptr==s) {
        /* no chars read */
        return NULL;
    } else {
        *ptr='\0';
        return s;
    }
}

int file_error(buffered_file *fh)
{
    return raw_file_error(fh->fh);
}


int file_eof(buffered_file *fh)
{
    return buf_avail(fh)==0 && raw_file_eof(fh->fh);
}
